--
-- pbLab-MessOfen Metatable Example V0.6
-- ASkr 9/2010
--
--
--
-- METATABLE FUNCTIONS:
--	 tMeas:init        (internal)
--	 tMeas:rawRead     (internal)
--	 tMeas:rawWrite    (internal)
--
--	 tMeas:SetOffset   Set new offset value (calibration).
--	 tMeas:CalcScale   Calculate new scale value (calibration).
--	 tMeas:SetScale    Set new scale value (calibration).
--
-- Usually, you will only need these:
--   tMeas:Calib       Semi-Automatic calibration function
--   tMeas:ShowCalib   Print out calibration values to console
--	 tMeas:Read        Read sensor value(s)
--	 tMeas.New         Initialize new sensor
--
--
-- EXAMPLE USAGE:
--
--   -- Open module #8 on port #2 in high (43V) range (1).
--   -- If the function calls leaves a table in "Vin", the call and the
--   -- initialization was sucessful. Otherwise, an errocode (number <0) is returned.
--   s1=tMeas.New(2,8,1)
--
--   -- First time only: Perform a calibration
--   s1:Calib()
--   -- On your NXT, push "left" for offset calibration and follow the instructions (short cables).
--   -- After this, press "right" for scale compensation and follow the instructions.
--   -- You will need a known voltage source for this! For best, universal performance,
--   -- you should choose ~3/4 of the max range (~30V for high or ~2V for low range).
--   -- Press cancel/exit button when done.
--   -- Note: Always perform the offset calibration first because the scale calibration
--   -- procedure uses this value too!
--
--   -- Now, read values with:
--   Vin=s1:Read()
--   -- Or, as a quick example, just call s1:Calib(), without pressing any other button
--   -- than cancel/exit:
--   s1:Calib()
--



collectgarbage("collect")

--table				=nil
debug				=nil
nxt_misc		=nil
nxt_bt			=nil
--nxt_math		=nil
nxt_file		=nil
nxt_output	=nil
--nxt_display	=nil
nxt_i2c			=nil
nxt_rs485		=nil
nxt_input		=nil
nxt_sound		=nil

collectgarbage("collect")

-- 18a:
nxt=nil
collectgarbage("collect")
require("nxt_misc")
require("nxt_input")
require("nxt_math") 
require("nxt_i2c")
require("nxt_file")
collectgarbage("collect")
-- (*2*) see above



------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- MODULE DEPENDENT CALIBRATION VALUES
-- 
-- These values will be used for offset and scale compensation.
-- The index of tCal[n], specifies the "sensor number" (*1*).
-- These values are changed by tMeas:SetOffset() and tMeas:SetScale().

--
-- Note: Offsets are integers (signed A/D values), scale values are pbLua floats!
-- Every table entry of tCal (usually tCal[1]..tCal[n]) contains a 4 number table:
--
--   tCal[n][1]  offset for low  range (2.25V)
--   tCal[n][2]  scale  for low  range (2.25V)
--   tCal[n][3]  offset for high range (43.0V)
--   tCal[n][4]  scale  for high range (43.0V)
--
-- (*1*) You are advised to label your modules!
--
-- The following tables contains the ideal scale values for low (1.09333)
-- and high range (21.0) and typical offset values.
-- NOTE: Different gain settings require re-calibration!
-- (e.g.: see sensor values for sensor number 8 below)
--
tCal={}
tCal[1]={4, nxt.float("1.093333"), 51, nxt.float("21.000000")}
tCal[2]={4, nxt.float("1.093333"), 51, nxt.float("21.000000")}
tCal[3]={4, nxt.float("1.093333"), 51, nxt.float("21.000000")}
tCal[4]={4, nxt.float("1.093333"), 51, nxt.float("21.000000")}
tCal[5]={4, nxt.float("1.093333"), 51, nxt.float("21.000000")}
tCal[6]={4, nxt.float("1.093333"), 51, nxt.float("21.000000")}
tCal[7]={4, nxt.float("1.093333"), 51, nxt.float("21.000000")}
tCal[8]={4, nxt.float("1.098841"), 51, nxt.float("21.593397")} -- gain=1     <- one of my modules
--tCal[8]={41, nxt.float("0.139920"), 349, nxt.float("3.235693")} -- gain=8  <- same module with gain==8


------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Some "defines"
-- 


-- I2C SPECIFIC
MCPADDR      =0xd0      -- address byte of MCP3421
cI2CTRIES    =   3      -- number of tries to communicate
cI2CREPDELAY = 100      -- delay, in ms, between multiple tries
cI2CTIMEOUT	 = 500      -- time in ms, after a comm error should bring up a timeout




--*********************
--*** DOWNLOAD STOP ***
--*********************
collectgarbage("collect")





--****************************************************************************
--****************************************************************************
--**** UTILITY FUNCTIONS AND METATABLES BELOW ********************************
--****************************************************************************
--****************************************************************************



--****************************************************************************
--** utilI2Comm
--**
--** <port>   -> port to communicate with
--** <sCmd>   -> table of values to send
--** <rBytes> -> number of bytes to receive ('0' if none)
--** 
--** return value(s):
--** timeout  -> nil
--** on error -> nil
--** OK       -> empty table {} (e.g.: write only)
--** answer   -> table of read bytes; { <b0>, <b1>, <b2>, ... }
--****************************************************************************
function utilI2Comm(port, sCmd, rBytes)
	local tim=nxt.TimerRead()
	local i,j
	local sTmp=""

	-- wait for I2C to become ready
	while ({nxt.I2CGetStatus(port)})[1] ~= 0 do
		if nxt.TimerRead() > tim+cI2CTIMEOUT then
			return nil
		end
	end

	sTmp=string.char(unpack(sCmd))

	tim=nxt.TimerRead()
	
	-- try to send or receive until timeout occurs
	for j=1,cI2CTRIES
	do
		-- send string
		nxt.I2CSendData(port,sTmp,rBytes)

		while 1 do
			local a,b,c=nxt.I2CGetStatus(port)

			if a==0 then
				if rBytes<1 then
					return {} -- no answer required
				else
					sTmp=nxt.I2CRecvData(port,rBytes)
					local tRet={}
					local i
					for i=1,rBytes do
						table.insert(tRet,string.byte(sTmp,i))
					end
					return tRet
				end --else return data
			end -- end if all OK

			-- check for timeout
			if nxt.TimerRead() > tim+cI2CTIMEOUT	then
				return nil --timeout
			end

			-- check for error (e.g.: missing ACKs...)
			if b==255 then
				utilWait(cI2CREPDELAY)
				break
			end

		end -- while status pending

	end -- while forever

	return nil

end -- function



--***************************************************************************
--** utilReadLine()
--**
--** Reads a string from the console.
--** <ENTER> accepts and returns the string
--** <ESC> aborts (empty string)
--***************************************************************************
function utilReadLine(msg)
	local ch
	local str=""

	nxt.FileWrite(0,msg)
	
	while 1 do
		ch=nxt.FileRead(0,1)
--		ch=string.upper(ch)	
		nxt.FileWrite(0,ch)
		if string.byte(ch)==13 then
			return str
		elseif string.byte(ch)==27 then
			print()
			return ""
		else
			str=str..ch
		end
	end 
end	



--***************************************************************************
--** utilBin2Int(<{dat}>)
--**
--** Converts the 3 byte, 18 bit mode answer of the MCP3421 to an integer.
--** Error checks removed to save memory...
--***************************************************************************
function utilBin2Int(dat)
	local i=0
	local j

	-- just in case ;)
--	if type(dat)~="table" then
--		return nil
--	end

--	if resol<RESOL12 or resol>RESOL18 then
--		return nil
--	end

--	if resol==RESOL18 and table.getn(dat)<3 then
--			return nil
--	elseif table.getn(dat)<2 then
--		return nil
--	end

	i=nxt.band(dat[1],0x03)*2^16 + dat[2]*2^8 + dat[3]
	-- check for negative number
	if nxt.band(i,0x00020000)~=0 then
		i=-nxt.bxor(i-1,0x0003ffff)
	end
	
	return i

end



--****************************************************************************
--** utilWait(<tim>)
--**
--** Waits (blocks) for ~<tim> ms
--****************************************************************************
function utilWait(tim)
	local tim1=nxt.TimerRead()
	repeat
	until nxt.TimerRead() > tim+tim1
end




--*********************
--*** DOWNLOAD STOP ***
--*********************
collectgarbage("collect")



------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- METATABLE
--
-- Does not use that much error checks to save memory!
--
tMeas={}
tMeas.__index=tMeas

-- VARIABLES
tMeas.port=0               -- the port number for this sensor
tMeas.range=0              -- selected measurement range (0=2.25V; 1=43V)
tMeas.module=0             -- number of module (for calibration value table tCal)
tMeas.gain=0               -- selected MCP3421 gain



-- FUNCTIONS
------------------------------------------------------------------------------
---- tMeas:init (internal usage only)
----
---- Sets port pins to I2C mode and initializes the MCP3421, by writing
---- its configuration register.
---- Mode is set to 18 bit, continuous operation.
---- Gain settings are take from "self.gain", which was filled by
---- tMeas:New() before...
---- After programming, the contents of the cfg reg are read back.
---- Initialization fails if:
----  - a comm error occurs (missing ACKs)
----  - the read back value of cfg reg is not what we sent
---- ERRORS:
----    0 -> OK
----   <0 -> ERROR
------------------------------------------------------------------------------
function tMeas:init()
	local cfgreg=0x10  -- MCP conf reg. -> continuous mode
	local i

	nxt.InputSetType(self.port,10)
	nxt.InputSetDir(self.port,1,1)
	nxt.InputSetState(self.port,1,1)
	nxt.I2CInitPins(self.port)

	nxt.InputSetState(self.port,1,1)

	utilWait(50)

  -- set resolution, gain and program the MCP
	if self.gain==8 then
		i=3
	elseif self.gain==4 then
		i=2
	elseif self.gain==2 then
		i=1
	else
		i=0
	end

	-- check and correct invalid range settings
	if range~=0 then
		range=1
	end
	
	cfgreg=cfgreg+(0x0c)+i

	if self:rawWrite({MCPADDR,cfgreg})~=OK then
		return -1
	end

  -- read back contents of the cfg register and compare to sent values
	utilWait(50)
	i=self:rawRead()
	
	if i==nil then
		return -2
	end
	
	if nxt.band(cfgreg,0x0f)==nxt.band(i[4],0x0f) then
		return 0
	else
		return -3
	end

end



------------------------------------------------------------------------------
---- tMeas:Calib
----
------------------------------------------------------------------------------
function tMeas:Calib()
	local but
	local i,j

	-- The main loop.
	-- Does nothing else but showing the current measurement value on the console.
	-- Hitting <left> or <right> initializes the offset or scale calibration function.
	while(1) do
		print(({self:Read()})[1])

		but=nxt.ButtonRead()
		-- cancel button
		if but==1 then
			break
		-- left button		
		elseif but==4 then
			print("OFFSET (short cables); <orange> continues...")
			while nxt.ButtonRead()~=8 do
			end
			print("SAMPLING...")
			j=0
			for i=1,11 do
				j=j+({self:Read()})[3]
				print(j/i)
				utilWait(400)				
			end
			j=j/11
			self:SetOffset(j)			
			print("OFFSET: "..tostring(j))
		-- right button
		elseif but==2 then
			print("SCALE (known voltage); <orange> continues...")
			while nxt.ButtonRead()~=8 do
			end
			print("SAMPLING...")
			j=0
			for i=1,11 do
				j=j+({self:Read()})[2]
				print(j/i)
				utilWait(400)				
			end
			j=j/11
			i=utilReadLine("enter real voltage (float):")
			if i~="" then
				i=nxt.float(i)
				i=self:CalcScale(j,i)			
				self:SetScale(i)
				print("SCALE:",i)
			else
				print("SKIPPED...")
			end
		end
		
		utilWait(400)
	end -- end while(1)
	
end --end s(port)




------------------------------------------------------------------------------
---- tMeas:rawRead
----
---- Four bytes are read from the MCP3421 and returned in a table.
---- Because we "only" support 18 bit mode:
---- First 3 bytes contain raw A/D values, with LSB in b[3] and MSB in b[1],
---- b[4] contains the value of the configuration register
---- (read MCP3421 datasheet for any clarifications)
---- ERRORS:
-----   table with values -> OK
-----   nil               -> ERROR
------------------------------------------------------------------------------
function tMeas:rawRead()
	local i
	i=utilI2Comm(self.port, {MCPADDR}, 4)

	if type(i)=="table" and table.getn(i)==4 then
		return i
	else
		return nil
	end
end



------------------------------------------------------------------------------
---- tMeas:rawWrite
----
---- Raw I2C write (without answer).
---- The byte contents of the given table <tabdata> are sent to
---- the sensor (writing its configuration register).
---- ERRORS:
----   0 -> OK
----  <0 -> ERROR (see "defines")
------------------------------------------------------------------------------
function tMeas:rawWrite(tabdata)
	local i;
	i=utilI2Comm(self.port, tabdata, 0)
	if type(i)=="table" and table.getn(i)==0 then
		return OK
	else
		return -4
	end
end



------------------------------------------------------------------------------
---- tMeas:SetOffset
----
---- Sets the offset of the sensor (software calibration).
---- <offset> has to specified in raw A/D digits.
------------------------------------------------------------------------------
function tMeas:SetOffset(offset)
	tCal[self.module][1+(self.range*2)]=offset
end


------------------------------------------------------------------------------
---- tMeas:SetScale
----
---- Sets the scale of the sensor (software calibration).
---- <scale> is a pbLua float, previously calculated by tMeas:CalcScale
------------------------------------------------------------------------------
function tMeas:SetScale(scale)
	tCal[self.module][2+(self.range*2)]=scale
end


------------------------------------------------------------------------------
---- tMeas:CalcScale
----
---- Note: All incoming and outgoing types are pbLua floats!
---- Calculates the correction factor ("scale")
---- float mVal - externally measured value
---- float dVal - value measured by Messofen (offset applied; tMeas.scale=1.0!)
------------------------------------------------------------------------------
function tMeas:CalcScale(dVal,mVal)
	local i
	
	-- dangerous, no checks here (dVal==0)!
	i=mVal/dVal
	
	return i
end


------------------------------------------------------------------------------
---- tMeas:ShowValib
----
---- Print all module related calibration values to the console.
------------------------------------------------------------------------------
function tMeas:ShowCalib()
	print("LOW")
	print("  offset: "..tostring(tCal[self.module][1]))
	print("  scale : "..tostring(tCal[self.module][2]))
	print("HIGH")
	print("  offset: "..tostring(tCal[self.module][3]))
	print("  scale : "..tostring(tCal[self.module][4]))
end



------------------------------------------------------------------------------
---- tMeas:Read
----
---- Returns 6 measurement values (a,b,c,d,e)
----
----   float a -> voltage with offset and scale applied   (read this for real meas.)
----   float b -> voltage with only offset applied        (needed for scale calib.)
----   int   c -> raw integer digits                      (needed for offset calib.)
----   int   d -> the current offset
----   int   e -> the current gain (1,2,4,8)
----
---- After calibration is done, your main application just needs the first value, e.g.:
----  umeas = Sens1:Read()
------------------------------------------------------------------------------
function tMeas:Read()
	local num
	local fr,fu
	local tmp
	num=self:rawRead()
	num=utilBin2Int(num)
	
	if num~=nil then
		tmp=nxt.float(num-tCal[self.module][1+(self.range*2)],0)
		fu=(tmp*nxt.float(2,48000))/nxt.float(2^17)
		fr=(tmp*nxt.float(2,48000)*tCal[self.module][2+(self.range*2)])/nxt.float(2^17)
		return fr,fu,num,tCal[self.module][1+(self.range*2)],self.gain
	else
		return nil,nil,nil,nil,nil -- just in case ;)
	end
end


--*********************
--*** DOWNLOAD STOP ***
--*********************
collectgarbage("collect")


------------------------------------------------------------------------------
---- tMeas.New (<port>, <module>, <range>, <gain>)
----
---- This needs to be called once before any of the interfaces can be used.
---- <port>  : number of the sensor port (integer: 1..4) the module is attached to
---- <module>: number (label) of module; is used for calibration table tCal
---- <range> : 0 selects 2V range, any other the 43V range
---- <gain>  : MCP3221 gain settings for this sensor; integer: 1,2,4 or 8;
----           defaults to 1, if omitted
----
---- Usually, you won't need any other gain but 1 (default).
---- 
---- ERRORS:
----   metatable with sensor object -> OK
----   a number value <0            -> ERROR
----
---- Error checks removed to save memory!
------------------------------------------------------------------------------
function tMeas.New(port,module,range,gain)
	local tm={}
	local i
	setmetatable(tm,tMeas)
	
	if gain==nil then
		gain=1
	end
	
	tm.port=port
	tm.module=module
	tm.range=range
	tm.gain=gain

	i=tm:init()
	if i==0 then
		return tm
	else
		return i
	end
end



--*********************
--*** DOWNLOAD STOP ***
--*********************
collectgarbage("collect")


